'''Высокоуровневый API для доступа к Битрикс24'''

from contextlib import asynccontextmanager, contextmanager
from typing import Iterable, Union

from . import correct_asyncio
from .srh import ServerRequestHandler
from .user_request import (BatchUserRequest, CallUserRequest,
                           GetAllUserRequest, GetByIDUserRequest,
                           ListAndGetUserRequest)


class BitrixAbstract(object):

    def __init__(self, webhook: str, verbose: bool = True,
                 respect_velocity_policy: bool = False):
        '''
        Создает объект класса Bitrix.

        Параметры:
        - `webhook: str` - URL вебхука, полученного от сервера Битрикс
        - `verbose: bool = True` - показывать ли прогрессбар при выполнении
        запроса
        - `respect_velocity_policy: bool = False` - соблюдать ли политику
        Битрикса о скорости запросов
        '''

        self.srh = ServerRequestHandler(webhook, respect_velocity_policy)
        self.verbose = verbose


class Bitrix(BitrixAbstract):
    '''Клиент для запросов к серверу Битрикс24.'''

    def get_all(self, method: str, params: dict = None) -> Union[list, dict]:
        '''
        Получить полный список сущностей по запросу `method`.

        Под капотом использует параллельные запросы и автоматическое построение
        батчей, чтобы ускорить получение данных. Также самостоятельно
        обратывает постраничные ответы сервера, чтобы вернуть полный список.

        Параметры:
        - `method` - метод REST API для запроса к серверу
        - `params` - параметры для передачи методу. Используется именно тот
            формат, который указан в документации к REST API Битрикс24.
            `get_all()` не поддерживает параметры `start`, `limit` и `order`.

        Возвращает полный список сущностей, имеющихся на сервере,
        согласно заданным методу и параметрам.
        '''

        return self.srh.run(GetAllUserRequest(self, method, params).run())

    def get_by_ID(self, method: str, ID_list: Iterable,
                  ID_field_name: str = 'ID', params: dict = None) -> dict:
        '''
        Получить список сущностей по запросу method и списку ID.

        Используется для случаев, когда нужны не все сущности,
        имеющиеся в базе, а конкретный список поименованных ID.
        Например, все контакты, привязанные к сделкам.

        Параметры:
        - `method` - метод REST API для запроса к серверу
        - `ID_list` - список ID
        - `ID_field_name` - название поля, которое будет подаваться в запрос
        для каждого элемента ID_list
        - `params` - параметры для передачи методу. Используется именно тот
        формат, который указан в документации к REST API Битрикс24

        Возвращает словарь вида:
        ```
        {
            ID_1: <результат запроса 1>,
            ID_2: <результат запроса 2>,
            ...
        }
        ```

        Значением элемента словаря будет результат выполнения запроса
        относительно этого ID. Это может быть, например, список связанных
        сущностей или пустой список, если не найдено ни одной привязанной
        сущности.
        '''

        return self.srh.run(GetByIDUserRequest(
            self, method, params, ID_list, ID_field_name).run())

    def list_and_get(self, method_branch: str, ID_field_name='ID') -> dict:
        '''
        Скачать список всех ID при помощи метода *.list,
        а затем все элементы при помощи метода *.get.

        Подобный подход показывает на порядок большую скорость
        получения данных, чем `get_all()`
        с параметром `'select': ['*', 'UF_*']`.

        Параметры:
        * `method_branch: str` - группа методов к использованию, например,
        `crm.lead` или `tasks.task`
        * `ID_field_name='ID'` - имя поля, в котором метод *.get принимает
        идентификаторы элементов (например, `'ID'` для метода `crm.lead.get`)

        Возвращает полное содержимое всех элементов в виде, используемом
        функцией `get_by_ID()` - словарь следующего вида:
        ```
        {
            ID_1: <словарь полей сущности с ID_1>,
            ID_2: <словарь полей сущности с ID_2>,
            ...
        }
        ```
        '''

        return self.srh.run(ListAndGetUserRequest(
            self, method_branch, ID_field_name).run())

    def call(self, method: str, items: Union[dict, Iterable]):
        '''
        Вызвать метод REST API по списку элементов.

        Параметры:
        - `method` - метод REST API
        - `items` - список параметров вызываемого метода
            либо dict с параметрами для единичного вызова

        Возвращает список ответов сервера для каждого из элементов `items`
        либо просто результат для единичного вызова.
        '''

        return self.srh.run(CallUserRequest(self, method, items).run())

    def call_batch(self, params: dict) -> dict:
        '''
        Вызвать метод `batch`.

        Параметры:
        - `params` - список параметров вызываемого метода

        Возвращает ответы сервера в формате словаря, где ключ - название
        команды, а значение - ответ сервера по этой команде.
        '''

        return self.srh.run(BatchUserRequest(self, params).run())

    @contextmanager
    def slow(self, max_concurrent_requests: int = 1):
        '''Временно ограничивает количество одновременно выполняемых запросов
        к Битрикс24.'''

        if not isinstance(max_concurrent_requests, int):
            raise 'slow() argument should be only int'
        if max_concurrent_requests < 1:
            raise 'slow() argument should be >= 1'

        mcr_max_backup, self.srh.mcr_max = \
            self.srh.mcr_max, max_concurrent_requests
        self.srh.mcr_cur_limit = min(self.srh.mcr_max, self.srh.mcr_cur_limit)

        yield True

        self.srh.mcr_max = mcr_max_backup
        self.srh.mcr_cur_limit = min(self.srh.mcr_max, self.srh.mcr_cur_limit)


class BitrixAsync(BitrixAbstract):
    '''Класс, повторяющий интерфейс класса `Bitrix`,
    но с асинхронными методами.'''

    async def get_all(self, method: str, params: dict = None) -> \
            Union[list, dict]:
        '''
        Получить полный список сущностей по запросу `method`.

        Под капотом использует параллельные запросы и автоматическое построение
        батчей, чтобы ускорить получение данных. Также самостоятельно
        обратывает постраничные ответы сервера, чтобы вернуть полный список.

        Параметры:
        - `method` - метод REST API для запроса к серверу
        - `params` - параметры для передачи методу. Используется
            именно тот формат, который указан в документации к REST API
            Битрикс24. `get_all()` не поддерживает параметры
            `start`, `limit` и `order`.

        Возвращает полный список сущностей, имеющихся на сервере,
        согласно заданным методу и параметрам.
        '''

        return await self.srh.run_async(
            GetAllUserRequest(self, method, params).run())

    async def get_by_ID(self, method: str, ID_list: Iterable,
                        ID_field_name: str = 'ID',
                        params: dict = None) -> list:
        '''
        Получить список сущностей по запросу `method` и списку ID.

        Используется для случаев, когда нужны не все сущности,
        имеющиеся в базе, а конкретный список поименованных ID.
        Например, все контакты, привязанные к сделкам.

        Параметры:
        - `method` - метод REST API для запроса к серверу
        - `ID_list` - список ID
        - `ID_field_name` - название поля, которое будет подаваться в запрос
        для каждого элемента ID_list
        - `params` - параметры для передачи методу. Используется именно тот
        формат, который указан в документации к REST API Битрикс24

        Возвращает словарь вида:
        ```
        {
            ID_1: <результат запроса 1>,
            ID_2: <результат запроса 2>,
            ...
        }
        ```

        Значением элемента словаря будет результат выполнения запроса
        относительно этого ID. Это может быть, например, список связанных
        сущностей или пустой список, если не найдено ни одной привязанной
        сущности.
        '''

        return await self.srh.run_async(GetByIDUserRequest(
            self, method, params, ID_list, ID_field_name).run())

    async def list_and_get(self, method_branch: str,
                           ID_field_name='ID') -> dict:
        '''
        Скачать список всех ID при помощи метода *.list,
        а затем все элементы при помощи метода *.get.

        Подобный подход показывает на порядок большую скорость
        получения данных, чем `get_all()`
        с параметром `'select': ['*', 'UF_*']`.

        Параметры:
        * `method_branch: str` - группа методов к использованию, например,
        `crm.lead` или `tasks.task`
        * `ID_field_name='ID'` - имя поля, в котором метод *.get принимает
        идентификаторы элементов (например, `'ID'` для метода `crm.lead.get`)

        Возвращает полное содержимое всех элементов в виде, используемом
        функцией `get_by_ID()` - словарь следующего вида:
        ```
        {
            ID_1: <словарь полей сущности с ID_1>,
            ID_2: <словарь полей сущности с ID_2>,
            ...
        }
        ```
        '''

        return await ListAndGetUserRequest(
            self, method_branch, ID_field_name=ID_field_name).run()

    async def call(self, method: str, items: Union[dict, Iterable]):
        '''
        Вызвать метод REST API по списку элементов.

        Параметры:
        - `method` - метод REST API
        - `items` - список параметров вызываемого метода
            либо dict с параметрами для единичного вызова

        Возвращает список ответов сервера для каждого из элементов `items`
        либо просто результат для единичного вызова.
        '''

        return await self.srh.run_async(
            CallUserRequest(self, method, items).run())

    async def call_batch(self, params: dict) -> dict:
        '''
        Вызвать метод `batch`.

        Параметры:
        - `params` - список параметров вызываемого метода

        Возвращает ответы сервера в формате словаря, где ключ - название
        команды, а значение - ответ сервера по этой команде.
        '''

        return await self.srh.run_async(
            BatchUserRequest(self, params).run())

    @asynccontextmanager
    async def slow(self, max_concurrent_requests: int = 1):
        '''Временно ограничивает количество одновременно выполняемых запросов
        к Битрикс24.'''

        if not isinstance(max_concurrent_requests, int):
            raise 'slow() argument should be only int'
        if max_concurrent_requests < 1:
            raise 'slow() argument should be >= 1'

        mcr_max_backup, self.srh.mcr_max = \
            self.srh.mcr_max, max_concurrent_requests
        self.srh.mcr_cur_limit = min(self.srh.mcr_max, self.srh.mcr_cur_limit)

        yield True

        self.srh.mcr_max = mcr_max_backup
        self.srh.mcr_cur_limit = min(self.srh.mcr_max, self.srh.mcr_cur_limit)



from asyncio import FIRST_COMPLETED, ensure_future, wait
from itertools import chain

from more_itertools import chunked
from tqdm import tqdm

from .server_response import ServerResponse
from .srh import BITRIX_MAX_BATCH_SIZE, ServerRequestHandler
from .utils import http_build_query


class MultipleServerRequestHandler:

    def __init__(self, bitrix, method, item_list,
                 real_len=None, real_start=0, mute=False):
        self.bitrix = bitrix
        self.srh: ServerRequestHandler = bitrix.srh
        self.method = method
        self.item_list = item_list
        self.real_len = real_len
        self.real_start = real_start
        self.mute = mute

        self.results = []
        self.task_iterator = self.generate_a_task()
        self.tasks = set()

    def generate_a_task(self):
        '''Объединяем элементы item_list в батчи и по одной создаем и
        возвращаем задачи asyncio с запросами к серверу для каждого батча.'''

        batches = ({
            'halt': 0,
            'cmd': {
                self.batch_command_label(i, item):
                f'{self.method}?{http_build_query(item)}'
                for i, item in enumerate(next_batch)
            }}
            for next_batch in chunked(self.item_list, BITRIX_MAX_BATCH_SIZE))

        for batch in batches:
            yield ensure_future(self.srh.single_request('batch', batch))

    def batch_command_label(self, i, item):
        return f'cmd{i}'

    async def run(self):
        self.top_up_tasks()

        with self.get_pbar() as pbar:
            while self.tasks:
                done, self.tasks = await wait(self.tasks,
                                              return_when=FIRST_COMPLETED)
                extracted_len = self.process_done_tasks(done)
                pbar.update(extracted_len)
                self.top_up_tasks()

    #            self.pbar.set_postfix({
    #                'max. requests': self.srh.mcr_cur_limit,
    #                'requests': self.srh.concurrent_requests,
    #                'tasks': len(self.tasks)})

        return self.results

    def top_up_tasks(self):
        '''Добавляем в self.tasks столько задач, сколько свободных слотов для
        запросов есть сейчас в self.srh.'''

        to_add = max(
            int(self.srh.mcr_cur_limit) - self.srh.concurrent_requests, 0)
        for _ in range(to_add):
            try:
                self.tasks.add(next(self.task_iterator))
            except StopIteration:
                break

    def process_done_tasks(self, done):
        '''Извлечь результаты из списка законченных задач
        и вернуть кол-во извлеченных элементов.'''

        extracted_len = 0
        for done_task in done:
            batch_response = done_task.result()
            unwrapped_result = ServerResponse(batch_response.result).result
            extracted_len += self.extract_result_from_batch_response(
                unwrapped_result)

        return extracted_len

    def extract_result_from_batch_response(self, unwrapped_result):
        '''Добавляет `unwrapped_result` в `self.results` и возвращает
        длину добавленного списка результатов'''

        if not unwrapped_result:
            return 0

        result_list = list(unwrapped_result.values())
        if type(result_list[0]) == list:
            result_list = list(chain(*result_list))
        self.results.extend(result_list)
        return len(result_list)

    def get_pbar(self):
        '''Возвращает прогресс бар `tqdm()` или пустышку,
        если `self.bitrix.verbose is False`.'''

        if self.bitrix.verbose and not self.mute:
            return tqdm(total=(self.real_len or len(self.item_list)),
                        initial=self.real_start)
        else:
            return MutePBar()


class MutePBar():
    def __enter__(self):
        return self

    def __exit__(*args):
        pass

    def update(*args):
        pass


class MultipleServerRequestHandlerPreserveIDs(MultipleServerRequestHandler):

    def __init__(self, bitrix, method, item_list, ID_field):
        super().__init__(bitrix, method, item_list)
        self.ID_field = ID_field
        self.results = {}

    def batch_command_label(self, i, item):
        return item[self.ID_field]

    def extract_result_from_batch_response(self, unwrapped_result):
        self.results.update(unwrapped_result)
        return len(unwrapped_result)



import asyncio
from asyncio import sleep, Event
import time
from collections import deque
from contextlib import asynccontextmanager

import aiohttp
from aiohttp.client_exceptions import (ClientPayloadError, ClientResponseError,
                                       ServerDisconnectedError)
from tqdm import tqdm

from .server_response import ServerResponse
from .utils import _url_valid


BITRIX_POOL_SIZE = 50
BITRIX_RPS = 2.0
BITRIX_MAX_BATCH_SIZE = 50
BITRIX_MAX_CONCURRENT_REQUESTS = 50

MAX_RETRIES = 10

RESTORE_CONNECTIONS_FACTOR = 1.3  # скорость восстановления количества запросов
DECREASE_CONNECTIONS_FACTOR = 3  # скорость уменьшения количества запросов
INITIAL_TIMEOUT = 0.5  # начальный таймаут в секундах
BACKOFF_FACTOR = 1.5  # основа расчета таймаута
# количество ошибок, до достижения котрого таймауты не делаются
NUM_FAILURES_NO_TIMEOUT = 3


class ServerError(Exception):
    pass


class ServerRequestHandler():
    '''
    Используется для контроля скорости доступа к серверам Битрикс.

    Основная цель - вести учет количества запросов, которые можно передать
    серверу Битрикс без получения ошибки `5XX`.

    Используется как контекстный менеджер, оборачивающий несколько
    последовательных запросов к серверу.
    '''

    def __init__(self, webhook, respect_velocity_policy):
        self.webhook = self.standardize_webhook(webhook)
        self.respect_velocity_policy = respect_velocity_policy

        self.requests_per_second = BITRIX_RPS
        self.pool_size = BITRIX_POOL_SIZE

        self.active_runs = 0
        self.session = None

        # rr - requests register - список отправленных запросов к серверу
        self.rr = deque()

        # лимит количества одновременных запросов,
        # установленный конструктором или пользователем
        self.mcr_max = BITRIX_MAX_CONCURRENT_REQUESTS

        # временный лимит количества одновременных запросов,
        # установленный через autothrottling
        self.mcr_cur_limit = BITRIX_MAX_CONCURRENT_REQUESTS

        self.concurrent_requests = 0
        self.request_complete = Event()

        # если положительное - количество последовательных удачных запросов
        # если отрицательное - количество последовательно полученных ошибок
        self.successive_results = 0

    @staticmethod
    def standardize_webhook(webhook):
        '''Приводит `webhook` к стандартному виду.'''

        if not isinstance(webhook, str):
            raise TypeError(f'Webhook should be a {str}')

        webhook = webhook.strip()

        if not _url_valid(webhook):
            raise ValueError('Webhook is not a valid URL')

        if webhook[-1] != '/':
            webhook += '/'

        return webhook

    def run(self, coroutine):
        '''Запускает `coroutine`, оборачивая его в `run_async()`.'''

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            asyncio.set_event_loop(asyncio.new_event_loop())
            loop = asyncio.get_event_loop()

        return loop.run_until_complete(self.run_async(coroutine))

    async def run_async(self, coroutine):
        '''Запускает `coroutine`, создавая и прекращая сессию
        при необходимости.'''

        async with self.handle_sessions():
            return await coroutine

    @asynccontextmanager
    async def handle_sessions(self):
        '''Открывает и закрывает сессию в зависимости от наличия
        активных запросов.'''

        if not self.active_runs and (not self.session or self.session.closed):
            self.session = aiohttp.ClientSession(raise_for_status=True)
        self.active_runs += 1

        try:
            yield True

        finally:
            self.active_runs -= 1
            if not self.active_runs and self.session and \
                    not self.session.closed:
                await self.session.close()

    async def single_request(self, method, params=None):
        '''Делает единичный запрос к серверу,
        с повторными попытками при необходимости.'''

        while True:

            try:
                result = await self.request_attempt(method, params)
                self.success()
                return result

            except (ClientPayloadError,
                    ServerDisconnectedError, ServerError) as err:
                self.failure(err)

    async def request_attempt(self, method, params=None):
        '''Делает попытку запроса к серверу, ожидая при необходимости.'''

        try:
            async with self.acquire(), self.session.post(
                    url=self.webhook + method, json=params) as response:
                return ServerResponse(await response.json(encoding='utf-8'))

        except ClientResponseError as error:
            if error.status // 100 == 5:  # ошибки вида 5XX
                raise ServerError('The server returned an error') from error
            else:
                raise

    def success(self):
        '''Увеличить счетчик удачных попыток.'''

        self.successive_results = max(self.successive_results + 1, 1)

    def failure(self, err: Exception):
        '''Увеличить счетчик неудачных попыток и поднять исключение,
        если попытки исчерпаны.'''

        self.successive_results = min(self.successive_results - 1, -1)

        if self.successive_results < -MAX_RETRIES:
            raise RuntimeError(
                'All attempts to get data from server exhausted') from err

    @asynccontextmanager
    async def acquire(self):
        '''Ожидает, пока не станет безопасно делать запрос к серверу.'''

        await self.autothrottle()

        async with self.limit_concurrent_requests():
            if self.respect_velocity_policy:
                async with self.limit_request_velocity():
                    yield
            else:
                yield

    async def autothrottle(self):
        '''Если было несколько неудач, делаем таймаут и уменьшаем скорость
        и количество одновременных запросов, и наоборот.'''

        if self.successive_results < 0:

            self.mcr_cur_limit = max(
                self.mcr_cur_limit / DECREASE_CONNECTIONS_FACTOR, 1)

            if self.successive_results < NUM_FAILURES_NO_TIMEOUT:
                power = -self.successive_results - NUM_FAILURES_NO_TIMEOUT - 1
                await sleep(INITIAL_TIMEOUT * BACKOFF_FACTOR ** power)

        elif self.successive_results > 0:

            self.mcr_cur_limit = min(
                self.mcr_cur_limit * RESTORE_CONNECTIONS_FACTOR, self.mcr_max)

    @asynccontextmanager
    async def limit_concurrent_requests(self):
        '''Не позволяет оновременно выполнять
        более `self.mcr_cur_limit` запросов.'''

        while self.concurrent_requests > self.mcr_cur_limit:
            self.request_complete.clear()
            await self.request_complete.wait()

        self.concurrent_requests += 1

        try:
            yield

        finally:
            self.concurrent_requests -= 1
            self.request_complete.set()

    @asynccontextmanager
    async def limit_request_velocity(self):
        '''Ограничивает скорость запросов к серверу.'''

        # если пул заполнен, ждать
        while len(self.rr) >= self.pool_size:
            time_from_last_request = time.monotonic() - self.rr[0]
            time_to_wait = 1 / self.requests_per_second - \
                time_from_last_request
            if time_to_wait > 0:
                await sleep(time_to_wait)
            else:
                break

        # зарегистрировать запрос в очереди
        start_time = time.monotonic()
        self.rr.appendleft(start_time)

        # отдать управление
        try:
            yield

        # подчистить пул
        finally:
            trim_time = start_time - self.pool_size / self.requests_per_second
            while self.rr[-1] < trim_time:
                self.rr.pop()



class ServerResponse():

    def __init__(self, response):
        self.response = response
        self.check_for_errors()

    def check_for_errors(self):
        if self.result_error:
            raise RuntimeError(
                f'The server reply contained an error: {self.result_error}')

    def __getattr__(self, item):

        # если результат вызова содержит только словарь {'tasks': список},
        # то вернуть этот список.
        # См. https://github.com/leshchenko1979/fast_bitrix24/issues/132
        if item == 'result':

            # для небатчевых запросов
            try:
                task_list = self.response['result']['tasks']
                if isinstance(task_list, list):
                    return task_list
            except (KeyError, TypeError):
                pass

            # для батчей
            try:
                return {batch_ID: batch_result['tasks']
                        for batch_ID, batch_result
                        in self.response['result'].items()}
            except (KeyError, TypeError, AttributeError):
                pass

        if item in self.response:
            return self.response[item]

        else:
            return None

    def more_results_expected(self):
        return self.total and self.total > 50 and \
            self.total != len(self.result)



from functools import wraps
from time import sleep
from urllib.parse import quote, urlparse


def _url_valid(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc, result.path])
    except:
        return False


def http_build_query(params, convention="%s"):
    if len(params) == 0:
        return ""

    output = ""
    for key in params.keys():

        if type(params[key]) is dict:

            output += http_build_query(params[key], convention % key + "[%s]")

        elif type(params[key]) is list:

            new_params = {str(i): element for i, element
                          in enumerate(params[key])}

            output += http_build_query(
                new_params, convention % key + "[%s]")

        else:

            val = quote(str(params[key]))
            key = quote(key)
            output = output + convention % key + "=" + val + "&"

    return output
